#ifndef __MSR_H__
#define __MSR_H__


/* AMD Defined MSRs */
#define MSR_K7_HWCR			0xC0010015
#define MSR_K8_VM_HSAVE_PA		0xC0010117

/* AMD/K8 specific MSRs */ 
#define MSR_EFER 0xc0000080		/* extended feature register */

/* EFER bits: */ 
#define _EFER_SCE   0  /* SYSCALL/SYSRET */
#define _EFER_LME   8  /* Long mode enable */
#define _EFER_NX   11  /* No execute enable */
#define _EFER_SVME 12
#define EFER_SCE   ( 1 << _EFER_SCE )
#define EFER_LME   ( 1 << _EFER_LME )
#define EFER_NX    ( 1 << _EFER_NX )
#define EFER_SVME  ( 1 << _EFER_SVME )


#ifndef __ASSEMBLY__

static inline void 
cpuid ( int op, unsigned int *eax, unsigned int *ebx,
	unsigned int *ecx, unsigned int *edx )
{
	__asm__("cpuid"
		: "=a" (*eax),
		  "=b" (*ebx),
		  "=c" (*ecx),
		  "=d" (*edx)
		: "0" (op));
}

static inline unsigned int 
cpuid_eax(unsigned int op)
{
	unsigned int eax;

	__asm__("cpuid"
		: "=a" (eax)
		: "0" (op)
		: "bx", "cx", "dx");
	return eax;
}

static inline 
unsigned int cpuid_ebx(unsigned int op)
{
	unsigned int eax, ebx;
	
	__asm__("cpuid"
		: "=a" (eax), "=b" (ebx)
		: "0" (op)
		: "cx", "dx" );
	return ebx;
}

static inline 
unsigned int cpuid_ecx(unsigned int op)
{
	unsigned int eax, ecx;
	
	__asm__("cpuid"
		: "=a" (eax), "=c" (ecx)
		: "0" (op)
		: "bx", "dx" );
	return ecx;
}

static inline 
unsigned int cpuid_edx(unsigned int op)
{
	unsigned int eax, edx;
	
	__asm__("cpuid"
		: "=a" (eax), "=d" (edx)
		: "0" (op)
		: "bx", "cx");
	return edx;
}

#define rdmsr(msr,val1,val2) \
       __asm__ __volatile__("rdmsr" \
			    : "=a" (val1), "=d" (val2) \
			    : "c" (msr))

#define wrmsr(msr,val1,val2) \
     __asm__ __volatile__("wrmsr" \
			  : /* no outputs */ \
			  : "c" (msr), "a" (val1), "d" (val2))

#endif /* __ASSEMBLY__ */


#endif /* __MSR_H__ */
