
static void
domain_create_arch ( struct domain *d )
{
	l1_pgentry_t gdt_l1e;
	int vcpuid, pdpt_order, rc;
	int i;
	
	pdpt_order = get_order_from_bytes(PDPT_L1_ENTRIES * sizeof(l1_pgentry_t));
	d->arch.mm_perdomain_pt = alloc_xenheap_pages(pdpt_order);
	memset(d->arch.mm_perdomain_pt, 0, PAGE_SIZE << pdpt_order);
	
	/*
	 * Map Xen segments into every VCPU's GDT, irrespective of whether every
	 * VCPU will actually be used. This avoids an NMI race during context
	 * switch: if we take an interrupt after switching CR3 but before switching
	 * GDT, and the old VCPU# is invalid in the new domain, we would otherwise
	 * try to load CS from an invalid table.
	 */
	gdt_l1e = l1e_from_page(virt_to_page(gdt_table), PAGE_HYPERVISOR);
	for ( vcpuid = 0; vcpuid < MAX_VIRT_CPUS; vcpuid++ ) {
		d->arch.mm_perdomain_pt[((vcpuid << GDT_LDT_VCPU_SHIFT) + FIRST_RESERVED_GDT_PAGE)] = gdt_l1e;
	}
	
	d->arch.mm_perdomain_l2 = alloc_xenheap_page();
	d->arch.mm_perdomain_l3 = alloc_xenheap_page();
	if ( (d->arch.mm_perdomain_l2 == NULL) || (d->arch.mm_perdomain_l3 == NULL) ) {
		fatal_failure ( "" );
	}
	
	memset(d->arch.mm_perdomain_l2, 0, PAGE_SIZE);
	for ( i = 0; i < (1 << pdpt_order); i++ ) {
		d->arch.mm_perdomain_l2[l2_table_offset(PERDOMAIN_VIRT_START)+i] =
			l2e_from_page(virt_to_page(d->arch.mm_perdomain_pt)+i, __PAGE_HYPERVISOR);
	}
	
	memset(d->arch.mm_perdomain_l3, 0, PAGE_SIZE);
	d->arch.mm_perdomain_l3[l3_table_offset(PERDOMAIN_VIRT_START)] =
		l3e_from_page(virt_to_page(d->arch.mm_perdomain_l2),  __PAGE_HYPERVISOR);
		
	shadow_lock_init(d);
	INIT_LIST_HEAD(&d->arch.free_shadow_frames);
	
	{
		d->arch.ioport_caps = rangeset_new(d, "I/O Ports", RANGESETF_prettyprint_hex);
		
		d->shared_info = alloc_xenheap_page();
		
		rc = ptwr_init(d);
		
		memset ( d->shared_info, 0, PAGE_SIZE );
		share_xen_page_with_guest ( virt_to_page ( d->shared_info ), d, XENSHARE_writable );
	}
}

struct vm
vm_create ( const struct vm_setup_info *dsi )
{
	struct domain d;
//	struct vcpu *v;
	
	memset ( &d, 0, sizeof ( struct domain ) );

	// atomic_set(&d->refcnt, 1);
	// spin_lock_init(&d->big_lock);
	// spin_lock_init(&d->page_alloc_lock);
	// INIT_LIST_HEAD(&d->page_list);
	// INIT_LIST_HEAD(&d->xenpage_list);
	
	// rangeset_domain_initialise(d);
	
//	{
//		set_bit(_DOMF_ctrl_pause, &d->domain_flags);
//		evtchn_init(d);
//		grant_table_create(d);
//	}
	
	domain_create_arch ( d );
	
//	d->iomem_caps = rangeset_new ( d, "I/O Memory", RANGESETF_prettyprint_hex );
//	d->irq_caps   = rangeset_new ( d, "Interrupts", 0 );
//	if ( (d->iomem_caps == NULL) || (d->irq_caps == NULL) ) {
//		fatal_failure ( "" ); /* NB. alloc_vcpu() is undone in free_domain() */
//	}

//	{
//		struct domain **pd;
//		write_lock(&domlist_lock);
//		pd = &domain_list; /* NB. domain_list maintained in order of dom_id. */
//		for ( pd = &domain_list; *pd != NULL; pd = &(*pd)->next_in_list ) {
//			if ( (*pd)->domain_id > d->domain_id ) {
//				break;
//			}
//		}
//		d->next_in_list = *pd;
//		*pd = d;
//		d->next_in_hashbucket = domain_hash[DOMAIN_HASH(dom_id)];
//		domain_hash[DOMAIN_HASH(dom_id)] = d;
//		write_unlock(&domlist_lock);
//	}

	construct_dom0 ( d, guest_kernel_start, guest_kernel_len );
	
	return d;
}

/*************************************************************/

void
write_ptbase ( struct vcpu *v )
{
	write_cr3 ( pagetable_get_paddr ( v->arch.monitor_table ) );
}

void discard_initial_images ( void )
{
	init_domheap_pages(initial_images_start, initial_images_end);
}

unsigned long 
NR ( unsigned long _l, unsigned long _h, unsigned long _s)
{
	return ( ( ( ( ( _h ) + ( ( 1UL << ( _s ) ) - 1 ) ) & 
		     ~ ( ( 1UL << ( _s ) ) - 1 ) ) - ( ( _l ) &
						       ~( ( 1UL << ( _s ) ) - 1 ) ) ) >> ( _s ) );
}


int
construct_dom0 ( struct domain *dom, unsigned long guest_kernel_start, unsigned long guest_kernel_size ) 
{
	struct vcpu *v = d->vcpu[0];

	int i;
	unsigned long pfn, mfn;
	unsigned long alloc_spfn;
	unsigned long alloc_epfn;
	unsigned long count;

	char *p;

	{
		/*
		 * If domain 0 allocation isn't specified, reserve 1/16th of available
		 * memory for things like DMA buffers. This reservation is clamped to 
		 * a maximum of 128MB.
		 */
		if ( dom0_nrpages == 0 ) {
			dom0_nrpages = avail_domheap_pages ( ) + initial_images_nrpages ( );
			dom0_nrpages = min ( dom0_nrpages / 16, 128L << ( 20 - PAGE_SHIFT ) );
			dom0_nrpages = - dom0_nrpages;
		}
		
		unsigned long nr_pages;
		/* Negative memory specification means "all memory - specified amount". */
		nr_pages = ( dom0_nrpages < 0 ) ? ( avail_domheap_pages ( ) + initial_images_nrpages ( ) + dom0_nrpages ) : dom0_nrpages;
	}

	/* ݤΰĴ (ڡơ֥ʬ) */
	

	/*
	 * This fully describes the memory layout of the initial domain. All 
	 * *_start address are page-aligned, except v_start (and v_end) which are 
	 * superpage-aligned.
	 */
	unsigned long vphysmap_start, vphysmap_end;
	unsigned long vstack_start, vstack_end;
	unsigned long vpt_start, vpt_end;
	unsigned long v_end;

	/*
	 * Why do we need this? The number of page-table frames depends on the 
	 * size of the bootstrap address space. But the size of the address space 
	 * depends on the number of page-table frames (since each one is mapped 
	 * read-only). We have a pair of simultaneous equations in two unknowns, 
	 * which we solve by exhaustive search.
	 */
	vphysmap_start   = PAGE_UP ( dsi.v_end );
	vphysmap_end     = vphysmap_start + (nr_pages * sizeof(unsigned long));
	vpt_start        = vphysmap_end;

	unsigned long nr_pt_pages;
	for ( nr_pt_pages = 2; ; nr_pt_pages++ ) {
		vpt_end          = vpt_start + (nr_pt_pages * PAGE_SIZE);
		vstack_start     = vpt_end;
		vstack_end       = vstack_start + PAGE_SIZE;
		v_end            = ( vstack_end + ( 1UL << 22 ) - 1 ) & ~( ( 1UL << 22 ) - 1 );
		if ( ( v_end - vstack_end ) < ( 512UL << 10 ) ) {
			v_end += 1UL << 22; /* Add extra 4MB to get >= 512kB padding. */
		}
		if ( ( 1 + /* # L4 */
		       NR ( dsi.v_start, v_end, L4_PAGETABLE_SHIFT ) + /* # L3 */
		       NR ( dsi.v_start, v_end, L3_PAGETABLE_SHIFT ) + /* # L2 */
		       NR ( dsi.v_start, v_end, L2_PAGETABLE_SHIFT ) )  /* # L1 */
		     <= nr_pt_pages )
			break;
	}

	/*  OS  load 뤿ΰݤ */

	order = get_order_from_bytes(v_end - dsi.v_start);
	/* Allocate from DMA pool: PAE L3 table must be below 4GB boundary. */
	struct page_info *page = alloc_domheap_pages ( d, /* order */, ALLOC_DOM_DMA );
	alloc_spfn = page_to_mfn ( page );
	alloc_epfn = alloc_spfn + d->tot_pages;

#if 0
	printf ( "PHYSICAL MEMORY ARRANGEMENT:\n" " Dom0 alloc.:   %x->%x", pfn_to_paddr ( alloc_spfn ), pfn_to_paddr ( alloc_epfn ) );
	printf ( "\nVIRTUAL MEMORY ARRANGEMENT:\n"
		 " Loaded kernel: %x->%x\n"
		 " Phys-Mach map: %x->%x\n"
		 " Page tables:   %x->%x\n"
		 " Boot stack:    %x->%x\n"
		 " TOTAL:         %x->%x\n",
		 dsi.v_kernstart, dsi.v_kernend, 
		 vphysmap_start, vphysmap_end,
		 vpt_start, vpt_end,
		 vstack_start, vstack_end,
		 dsi.v_start, v_end );
	printf ( " ENTRY ADDRESS: %x\n", dsi.v_kernentry );
#endif

	/* Machine address of next candidate page-table page. */
	unsigned long mpt_alloc = ( vpt_start - dsi.v_start ) + ( unsigned long ) pfn_to_paddr ( alloc_spfn );

	{ 
		/* WARNING: The new domain must have its 'processor' field filled in! */
		l4_pgentry_t *l4tab = NULL, *l4start = NULL;
		l3_pgentry_t *l3tab = NULL, *l3start = NULL;
		l2_pgentry_t *l2tab = NULL, *l2start = NULL;
		l1_pgentry_t *l1tab = NULL, *l1start = NULL;

		maddr_to_page ( mpt_alloc )->u.inuse.type_info = PGT_l4_page_table;
		l4start = l4tab = VIRT ( mpt_alloc ); mpt_alloc += PAGE_SIZE;
		memcpy(l4tab, &idle_pg_table[0], PAGE_SIZE);
		l4tab[l4_table_offset(LINEAR_PT_VIRT_START)] = l4e_from_paddr(__pa(l4start), __PAGE_HYPERVISOR);
		l4tab[l4_table_offset(PERDOMAIN_VIRT_START)] = l4e_from_paddr(__pa(d->arch.mm_perdomain_l3), __PAGE_HYPERVISOR);
		v->arch.guest_table = mk_pagetable ( __pa ( l4start ) );

		l4tab += l4_table_offset(dsi.v_start);
		mfn = alloc_spfn;
		for ( count = 0; count < ( ( v_end - dsi.v_start ) >> PAGE_SHIFT ); count++ ) {
			if ( ! ( ( unsigned long ) l1tab & ( PAGE_SIZE - 1 ) ) ) {
				maddr_to_page(mpt_alloc)->u.inuse.type_info = PGT_l1_page_table;
				l1start = l1tab = __va(mpt_alloc); mpt_alloc += PAGE_SIZE;
				clear_page(l1tab);
				if ( count == 0 ) {
					l1tab += l1_table_offset(dsi.v_start);
				}
				if ( ! ( ( unsigned long ) l2tab & ( PAGE_SIZE - 1 ) ) ) {
					maddr_to_page(mpt_alloc)->u.inuse.type_info = PGT_l2_page_table;
					l2start = l2tab = __va(mpt_alloc); mpt_alloc += PAGE_SIZE;
					clear_page(l2tab);
					if ( count == 0 ) {
						l2tab += l2_table_offset(dsi.v_start);
					}
					if ( ! ( ( unsigned long ) l3tab & ( PAGE_SIZE - 1 ) ) ) {
						maddr_to_page(mpt_alloc)->u.inuse.type_info =
							PGT_l3_page_table;
						l3start = l3tab = __va(mpt_alloc); mpt_alloc += PAGE_SIZE;
						clear_page(l3tab);
						if ( count == 0 )
							l3tab += l3_table_offset(dsi.v_start);
						*l4tab = l4e_from_paddr(__pa(l3start), L4_PROT);
						l4tab++;
					}
					*l3tab = l3e_from_paddr(__pa(l2start), L3_PROT);
					l3tab++;
				}
				*l2tab = l2e_from_paddr(__pa(l1start), L2_PROT);
				l2tab++;
			}
			*l1tab = l1e_from_pfn(mfn, L1_PROT);
			l1tab++;

			struct page_info *page = NULL;
			page = mfn_to_page ( mfn );
			if ( ( page->u.inuse.type_info == 0 ) && !get_page_and_type ( page, d, PGT_writable_page ) ) {
				BUG();
			}

			mfn++;
		}

		/* Pages that are part of page tables must be read only. */
		l4tab = l4start + l4_table_offset(vpt_start);
		l3start = l3tab = l4e_to_l3e(*l4tab);
		l3tab += l3_table_offset(vpt_start);
		l2start = l2tab = l3e_to_l2e(*l3tab);
		l2tab += l2_table_offset(vpt_start);
		l1start = l1tab = l2e_to_l1e(*l2tab);
		l1tab += l1_table_offset(vpt_start);
		for ( count = 0; count < nr_pt_pages; count++ ) {
			struct page_info *page = NULL;

			l1e_remove_flags(*l1tab, _PAGE_RW);
			page = mfn_to_page(l1e_get_pfn(*l1tab));

			/* Read-only mapping + PGC_allocated + page-table page. */
			page->count_info         = PGC_allocated | 3;
			page->u.inuse.type_info |= PGT_validated | 1;

			/* Top-level p.t. is pinned. */
			if ( ( page->u.inuse.type_info & PGT_type_mask ) == PGT_l4_page_table ) {
				page->count_info        += 1;
				page->u.inuse.type_info += 1 | PGT_pinned;
			}

			/* Iterate. */
			if ( !((unsigned long)++l1tab & (PAGE_SIZE - 1)) ) {
				if ( !((unsigned long)++l2tab & (PAGE_SIZE - 1)) )
				{
					if ( !((unsigned long)++l3tab & (PAGE_SIZE - 1)) )
						l3start = l3tab = l4e_to_l3e(*++l4tab); 
					l2start = l2tab = l3e_to_l2e(*l3tab);
				}
				l1start = l1tab = l2e_to_l1e(*l2tab);
			}
		}
	}

	/*********************************************************/
	
	/* Set up monitor table */
	update_pagetables ( v ); 

	/* Install the new page tables. */
	write_ptbase ( v );

	/*
	  - guest OS (elf ) ɤ
	     - multiboot ǼơؤΥݥ󥿤 ebx 쥸ꤷƤɬפ롥

	     - elf header ˽äơʪ˥
	       - ʪɥ쥹Ѵɬ
	*/


	/* Free temporary buffers. */
	discard_initial_images ( ); // ???


	/* VM's physical --> PM's physical Ρ˥ڡơ֥򹹿 */


	if ( opt_dom0_shadow ) {
		shadow_mode_enable ( d, SHM_enable );
		update_pagetables ( v );
	}

	/*
	 * We're basically forcing default RPLs to 1, so that our "what privilege
	 * level are we returning to?" logic works.
	 */
	v->arch.guest_context.kernel_ss = FLAT_KERNEL_SS;
	for ( i = 0; i < 256; i++ ) {
		v->arch.guest_context.trap_ctxt [ i ].cs = FLAT_KERNEL_CS;
	}

#ifdef SUPERVISOR_MODE_KERNEL
	v->arch.guest_context.kernel_ss &= ~3;
	v->arch.guest_context.user_regs.ss &= ~3;
	v->arch.guest_context.user_regs.es &= ~3;
	v->arch.guest_context.user_regs.ds &= ~3;
	v->arch.guest_context.user_regs.fs &= ~3;
	v->arch.guest_context.user_regs.gs &= ~3;
	printk("Dom0 runs in ring 0 (supervisor mode)\n");
#endif /* SUPERVISOR_MODE_KERNEL */
}
