long
set_gdt ( struct vcpu *v, unsigned long *frames, unsigned int entries )
{
	struct domain *d = v->domain;
	/* NB. There are 512 8-byte entries per GDT page. */
	int i, nr_pages = (entries + 511) / 512;
	unsigned long mfn;
	
	if ( entries > FIRST_RESERVED_GDT_ENTRY ) {
		return -EINVAL;
	}

	shadow_sync_all(d);

	/* Check the pages in the new GDT. */
	for ( i = 0; i < nr_pages; i++ ) {
		mfn = frames[i] = gmfn_to_mfn(d, frames[i]);
		if ( !mfn_valid(mfn) || !get_page_and_type ( mfn_to_page ( mfn ), d, PGT_gdt_page ) ) {
			goto fail;
		}
	}

	/* Tear down the old GDT. */
	destroy_gdt ( v );

	/* Install the new GDT. */
	v->arch.guest_context.gdt_ents = entries;
	for ( i = 0; i < nr_pages; i++ ) {
		v->arch.guest_context.gdt_frames[i] = frames[i];
		v->arch.perdomain_ptes[i] = l1e_from_pfn ( frames[i], __PAGE_HYPERVISOR );
	}

	return 0;

fail:
	while ( i-- > 0 )
		put_page_and_type(mfn_to_page(frames[i]));
	return -EINVAL;
}


