#include "types.h"
#include "bitops.h"
#include "string.h"
#include "printf.h"
#include "failure.h"
#include "page.h"
#include "msr.h"
#include "processor.h"
#include "cpufeature.h"
#include "svm.h"
#include "alloc.h"

/* AMD64 manual Vol. 2, p. 441 */
/* Host save area */
static void *host_save_area;


/* ҡפ˳ݤɬפ롩 */
void *
alloc_host_save_area ( void )
{
	void *hsa;

	unsigned long n  = alloc_pages ( 1, 1 );
	hsa = ( void * ) VIRT ( n << PAGE_SHIFT );
	
	if ( hsa ) {
		memset ( hsa, 0, PAGE_SIZE );
	}
	
	return hsa;
}

void __init
enable_svm ( struct cpuinfo_x86 *c )
{
 	/* Xen does not fill x86_capability words except 0. */
	{
		u32 ecx = cpuid_ecx ( 0x80000001 );
		c->x86_capability[5] = ecx;
	}
    
	if ( ! ( test_bit ( X86_FEATURE_SVME, &c->x86_capability ) ) ) {
		fatal_failure ( "No svm featurel\n" );
		return;
	}
   
	{ /* Before any SVM instruction can be used, EFER.SVME (bit 12
	   * of the EFER MSR register) must be set to 1.  
	   * (See AMD64 manual Vol. 2, p. 439) */
		u32 eax, edx;
		rdmsr ( MSR_EFER, eax, edx );
		eax |= EFER_SVME;
		wrmsr ( MSR_EFER, eax, edx );
	}

	printf ( "AMD SVM Extension is enabled.\n" );

	/* Initialize the HSA */
	{
		u64 phys_hsa;
		u32 phys_hsa_lo, phys_hsa_hi;   

		host_save_area = alloc_host_save_area ( );
		phys_hsa = ( u64 ) PHYS ( host_save_area ); 
		phys_hsa_lo = ( u32 ) phys_hsa;
		phys_hsa_hi = ( u32 ) (phys_hsa >> 32);    
		wrmsr ( MSR_K8_VM_HSAVE_PA, phys_hsa_lo, phys_hsa_hi ); 
	}
}


/*****************************************************************************************************/

#if 0

void 
svm_vmexit_handler ( struct cpu_user_regs regs )
{
	unsigned int exit_reason;
	unsigned long eip;
	struct vcpu *v = current;
	int error;
	int do_debug = 0;
	struct vmcb_struct *vmcb = v->arch.hvm_svm.vmcb;

	exit_reason = vmcb->exitcode;
	save_svm_cpu_user_regs(v, &regs);
	v->arch.hvm_svm.injecting_event = 0;

	vmcb->tlb_control = 1;

	if (exit_reason == -1)
	{
		printk("%s: exit_reason == -1 - Did someone clobber the VMCB\n", 
		       __func__);
		BUG();
		domain_crash_synchronous();
	}

	perfc_incra(vmexits, exit_reason);
	eip = vmcb->rip;


	TRACE_3D(TRC_VMX_VMEXIT, v->domain->domain_id, eip, exit_reason);

	switch (exit_reason) 
	{
	case VMEXIT_EXCEPTION_DB:
	{
		svm_store_cpu_user_regs(&regs, v);
		domain_pause_for_debugger();  
	}
	break;

	case VMEXIT_NMI:
		do_nmi(&regs, 0);
		break;

	case VMEXIT_SMI:
		/*
		 * For asynchronous SMI's, we just need to allow global interrupts 
		 * so that the SMI is taken properly in the context of the host.  The
		 * standard code does a STGI after the VMEXIT which should accomplish 
		 * this task.  Continue as normal and restart the guest.
		 */
		break;

	case VMEXIT_INIT:
		/*
		 * Nothing to do, in fact we should never get to this point. 
		 */
		break;

	case VMEXIT_EXCEPTION_BP:
		if ( test_bit(_DOMF_debugging, &v->domain->domain_flags) )
			domain_pause_for_debugger();
		else 
			svm_inject_exception(vmcb, TRAP_int3, 0, 0);
		break;

	case VMEXIT_EXCEPTION_NM:
		svm_do_no_device_fault(vmcb);
		break;  

	case VMEXIT_EXCEPTION_GP:
		/* This should probably not be trapped in the future */
		regs.error_code = vmcb->exitinfo1;
		v->arch.hvm_svm.injecting_event = 1;
		svm_do_general_protection_fault(v, &regs);
		break;  

	case VMEXIT_EXCEPTION_PF:
	{
		unsigned long va;
		va = vmcb->exitinfo2;
		regs.error_code = vmcb->exitinfo1;
		HVM_DBG_LOG(DBG_LEVEL_VMMU, 
			    "eax=%lx, ebx=%lx, ecx=%lx, edx=%lx, esi=%lx, edi=%lx",
			    (unsigned long)regs.eax, (unsigned long)regs.ebx,
			    (unsigned long)regs.ecx, (unsigned long)regs.edx,
			    (unsigned long)regs.esi, (unsigned long)regs.edi);

		v->arch.hvm_vcpu.mmio_op.inst_decoder_regs = &regs;

//printk("PF1\n");
		if (!(error = svm_do_page_fault(va, &regs))) 
		{
			v->arch.hvm_svm.injecting_event = 1;
			/* Inject #PG using Interruption-Information Fields */
			svm_inject_exception(vmcb, TRAP_page_fault, 1, regs.error_code);

			v->arch.hvm_svm.cpu_cr2 = va;
			vmcb->cr2 = va;
			TRACE_3D(TRC_VMX_INT, v->domain->domain_id, 
				 VMEXIT_EXCEPTION_PF, va);
		}
		break;
	}

	case VMEXIT_EXCEPTION_DF:
		printk("Guest double fault");
		BUG();
		break;

	case VMEXIT_INTR:
		raise_softirq(SCHEDULE_SOFTIRQ);
		break;

	case VMEXIT_GDTR_WRITE:
		printk("WRITE to GDTR\n");
		break;

	case VMEXIT_TASK_SWITCH:
		__hvm_bug(&regs);
		break;

	case VMEXIT_CPUID:
		svm_vmexit_do_cpuid(vmcb, regs.eax, &regs);
		break;

	case VMEXIT_HLT:
		svm_vmexit_do_hlt(vmcb);
		break;

	case VMEXIT_INVLPG:
		svm_handle_invlpg(0, &regs);
		break;

	case VMEXIT_INVLPGA:
		svm_handle_invlpg(1, &regs);
		break;

	case VMEXIT_VMMCALL:
		svm_do_vmmcall(v, &regs);
		break;

	case VMEXIT_CR0_READ:
		svm_cr_access(v, 0, TYPE_MOV_FROM_CR, &regs);
		break;

	case VMEXIT_CR2_READ:
		svm_cr_access(v, 2, TYPE_MOV_FROM_CR, &regs);
		break;

	case VMEXIT_CR3_READ:
		svm_cr_access(v, 3, TYPE_MOV_FROM_CR, &regs);
		break;

	case VMEXIT_CR4_READ:
		svm_cr_access(v, 4, TYPE_MOV_FROM_CR, &regs);
		break;

	case VMEXIT_CR8_READ:
		svm_cr_access(v, 8, TYPE_MOV_FROM_CR, &regs);
		break;

	case VMEXIT_CR0_WRITE:
		svm_cr_access(v, 0, TYPE_MOV_TO_CR, &regs);
		break;

	case VMEXIT_CR2_WRITE:
		svm_cr_access(v, 2, TYPE_MOV_TO_CR, &regs);
		break;

	case VMEXIT_CR3_WRITE:
		svm_cr_access(v, 3, TYPE_MOV_TO_CR, &regs);
		local_flush_tlb();
		break;

	case VMEXIT_CR4_WRITE:
		svm_cr_access(v, 4, TYPE_MOV_TO_CR, &regs);
		break;

	case VMEXIT_CR8_WRITE:
		svm_cr_access(v, 8, TYPE_MOV_TO_CR, &regs);
		break;

	case VMEXIT_DR0_READ:
		svm_dr_access(v, 0, TYPE_MOV_FROM_DR, &regs);
		break;

	case VMEXIT_DR1_READ:
		svm_dr_access(v, 1, TYPE_MOV_FROM_DR, &regs);
		break;

	case VMEXIT_DR2_READ:
		svm_dr_access(v, 2, TYPE_MOV_FROM_DR, &regs);
		break;

	case VMEXIT_DR3_READ:
		svm_dr_access(v, 3, TYPE_MOV_FROM_DR, &regs);
		break;

	case VMEXIT_DR6_READ:
		svm_dr_access(v, 6, TYPE_MOV_FROM_DR, &regs);
		break;

	case VMEXIT_DR7_READ:
		svm_dr_access(v, 7, TYPE_MOV_FROM_DR, &regs);
		break;

	case VMEXIT_DR0_WRITE:
		svm_dr_access(v, 0, TYPE_MOV_TO_DR, &regs);
		break;

	case VMEXIT_DR1_WRITE:
		svm_dr_access(v, 1, TYPE_MOV_TO_DR, &regs);
		break;

	case VMEXIT_DR2_WRITE:
		svm_dr_access(v, 2, TYPE_MOV_TO_DR, &regs);
		break;

	case VMEXIT_DR3_WRITE:
		svm_dr_access(v, 3, TYPE_MOV_TO_DR, &regs);
		break;

	case VMEXIT_DR6_WRITE:
		svm_dr_access(v, 6, TYPE_MOV_TO_DR, &regs);
		break;

	case VMEXIT_DR7_WRITE:
		svm_dr_access(v, 7, TYPE_MOV_TO_DR, &regs);
		break;

	case VMEXIT_IOIO:
		svm_io_instruction(v, &regs);
		break;

	case VMEXIT_MSR:
		svm_do_msr_access(v, &regs);
		break;

	case VMEXIT_SHUTDOWN:
		printk("Guest shutdown exit\n");
		domain_crash_synchronous();
		break;

	default:
		printk("unexpected VMEXIT: exit reason = 0x%x, exitinfo1 = %llx, "
		       "exitinfo2 = %llx\n", exit_reason, 
		       (unsigned long long)vmcb->exitinfo1, 
		       (unsigned long long)vmcb->exitinfo2);
		__hvm_bug(&regs);       /* should not happen */
		break;
	}

	return;
}

#endif


#if 0

/*****************************************************************************************************/

void 
svm_final_setup_guest ( struct vcpu *v )
{
	struct domain *d = v->domain;
	struct vcpu *vc;
	
	/* Initialize monitor page table */
	for_each_vcpu(d, vc)
		vc->arch.monitor_table = mk_pagetable(0);
	
	/* 
	 * Required to do this once per domain
	 * TODO: add a seperate function to do these.
	 */
	memset ( &d->shared_info->evtchn_mask[0], 0xff, sizeof ( d->shared_info->evtchn_mask ) );
	
	/* 
	 * Put the domain in shadow mode even though we're going to be using
	 * the shared 1:1 page table initially. It shouldn't hurt 
	 */
	shadow_mode_enable ( d, ( SHM_enable | SHM_refcounts | SHM_translate | SHM_external | SHM_wr_pt_pte ) );
}

/*****************************************************************************************************/



/* Make sure that xen intercepts any FP accesses from current */
static void 
svm_stts ( struct vcpu *v ) 
{
	/* FPU state already dirty? Then no need to setup_fpu() lazily. */
	if ( test_bit(_VCPUF_fpu_dirtied, &v->vcpu_flags) ) {
		return;
	}

	/*
	 * If the guest does not have TS enabled then we must cause and handle an 
	 * exception on first use of the FPU. If the guest *does* have TS enabled 
	 * then this is not necessary: no FPU activity can occur until the guest 
	 * clears CR0.TS, and we will initialise the FPU when that happens.
	 */
	if ( ! ( v->arch.hvm_svm.cpu_shadow_cr0 & X86_CR0_TS ) ) {
		v->arch.hvm_svm.vmcb->exception_intercepts |= EXCEPTION_BITMAP_NM;
		vmcb->cr0 |= X86_CR0_TS;
	}
}

/*****************************************************************************************************/

void
hvm_setup_platform ( struct domain* d )
{
	struct hvm_domain *platform;

	if ( ! hvm_guest ( current ) || ( current->vcpu_id != 0 ) ) {
		return;
	}

	shadow_direct_map_init(d);

	hvm_map_io_shared_page(d);
	hvm_get_info(d);

	platform = &d->arch.hvm_domain;
	pic_init(&platform->vpic, pic_irq_request, &platform->interrupt_request);
	register_pic_io_hook();

	if ( hvm_apic_support(d) ) {
		spin_lock_init(&d->arch.hvm_domain.round_robin_lock);
		hvm_vioapic_init(d);
	}
}

static int 
svm_paging_enabled ( struct vcpu *v )
{
	unsigned long cr0;

	cr0 = v->arch.hvm_svm.cpu_shadow_cr0;

	return ( cr0 & X86_CR0_PE ) && ( cr0 & X86_CR0_PG );
}

void 
svm_do_launch ( struct vcpu *v )
{
	struct vmcb_struct *vmcb = v->arch.hvm_svm.vmcb;
	int core = smp_processor_id();
	
	/* Update CR3, GDT, LDT, TR */
	svm_stts(v);
	
	/* current core is the one we intend to perform the VMRUN on */
//	v->arch.hvm_svm.launch_core = v->arch.hvm_svm.asid_core = core;
//	clear_bit ( ARCH_SVM_VMCB_ASSIGN_ASID, &v->arch.hvm_svm.flags );
//	asidpool_assign_next( vmcb, 0, core, core );

	hvm_setup_platform ( v->domain );
	
//	evtchn_bind_vcpu ( iopacket_port(v), v->vcpu_id );
//	clear_bit ( iopacket_port(v), &v->domain->shared_info->evtchn_mask[0]);
//	
//	if (hvm_apic_support(v->domain)) {
//		vlapic_init(v);
//	}

	init_timer(&v->arch.hvm_svm.hlt_timer,  hlt_timer_fn, v, v->processor);
	
	vmcb->ldtr.sel = 0;
	vmcb->ldtr.base = 0;
	vmcb->ldtr.limit = 0;
	vmcb->ldtr.attributes.bytes = 0;
	
	vmcb->efer = EFER_SVME; /* Make sure VMRUN won't return with -1 */

	vmcb->cr3 = pagetable_get_paddr ( svm_paging_enabled(v) ? v->arch.guest_table : ov->domain->arch.phys_table );

//	v->arch.hvm_svm.injecting_event  = 0;
	v->arch.hvm_svm.saved_irq_vector = -1;
	
	svm_set_guest_time ( v, 0 );
	
	vmcb->tlb_control = 1;
}

static void
arch_svm_do_launch(struct vcpu *v) 
{
	cpu_user_regs_t *regs = &current->arch.guest_context.user_regs;
	
	construct_vmcb ( &v->arch.hvm_svm, regs );
	
	svm_do_launch ( v );
	
	reset_stack_and_jump ( svm_asm_do_launch ); 
}

/*********************************************************************************************/

static void 
svm_freeze_time ( struct vcpu *v )
{
	struct hvm_virpit *vpit = &v->domain->arch.hvm_domain.vpit;
	
	if ( vpit->first_injected && !v->domain->arch.hvm_domain.guest_time ) {
		v->domain->arch.hvm_domain.guest_time = svm_get_guest_time(v);
		stop_timer(&(vpit->pit_timer));
	}
}

static void 
svm_ctxt_switch_from ( struct vcpu *v )
{
	svm_freeze_time(v);
}

/*********************************************************************************************/

static void 
svm_ctxt_switch_to ( struct vcpu *v )
{
	/* 
	 * This is required, because VMRUN does consistency check
	 * and some of the DOM0 selectors are pointing to 
	 * invalid GDT locations, and cause AMD processors
	 * to shutdown.
	 */
	set_segment_register(ds, 0);
	set_segment_register(es, 0);
	set_segment_register(ss, 0);
}

/*********************************************************************************************/

static void
svm_do_resume ( struct vcpu *v ) 
{
	struct domain *d = v->domain;
	struct hvm_virpit *vpit = &d->arch.hvm_domain.vpit;

	svm_stts(v);

	/* make sure the HSA is set for the current core */
	set_hsa_to_guest( &v->arch.hvm_svm );
    
	/* pick up the elapsed PIT ticks and re-enable pit_timer */
	if ( vpit->first_injected ) {
		if ( v->domain->arch.hvm_domain.guest_time ) {
			svm_set_guest_time(v, v->domain->arch.hvm_domain.guest_time);
			v->domain->arch.hvm_domain.guest_time = 0;
		}
		pickup_deactive_ticks(vpit);
	}

	if ( test_bit(iopacket_port(v), &d->shared_info->evtchn_pending[0]) || test_bit(ARCH_HVM_IO_WAIT, &v->arch.hvm_vcpu.ioflags) ) {
		hvm_wait_io();
	}
}

void
arch_svm_do_resume ( struct vcpu *v ) 
{
	svm_do_resume ( v );
	reset_stack_and_jump( svm_asm_do_resume );
}

#endif /* 0 */

