static struct vcpu *
vcpu_alloc ( struct domain *d )
{
	struct vcpu *v;
	
	v = xmalloc ( struct vcpu );
	memset ( v, 0, sizeof ( *v ) );
	
	return v;
}

struct vcpu *
vcpu_create ( void )
{
	struct vcpu *v;

	v = vcpu_alloc ( d );

	v->domain = NULL;
	
	v->arch.flags = TF_kernel_mode;
	
	v->arch.perdomain_ptes   = d->arch.mm_perdomain_pt + (vcpu_id << GDT_LDT_VCPU_SHIFT);
	
	v->arch.guest_vtable  = __linear_l2_table;
	v->arch.shadow_vtable = __shadow_linear_l2_table;
	v->arch.guest_vl3table = __linear_l3_table;
	v->arch.guest_vl4table = __linear_l4_table;

	atomic_set ( &v->pausecnt, 0 );
	v->vcpu_info = &d->shared_info->vcpu_info[vcpu_id];

//	v->cpu_affinity = CPU_MASK_ALL;

//	v->runstate.state = RUNSTATE_offline;
//	v->runstate.state_entry_time = NOW();

	if ( ( vcpu_id != 0 ) && ( ! is_idle_domain ( d ) ) ) {
		set_bit(_VCPUF_down, &v->vcpu_flags);
	}

	return v;
}

/*************************************************************************/


static int 
arch_set_info_guest ( struct vcpu *v, struct vcpu_guest_context *c )
{
	struct domain *d = v->domain;
	unsigned long phys_basetab = INVALID_MFN;
	int i, rc;

	if ( ! ( c->flags & VGCF_HVM_GUEST ) )
	{
		fixup_guest_stack_selector(c->user_regs.ss);
		fixup_guest_stack_selector(c->kernel_ss);
		fixup_guest_code_selector(c->user_regs.cs);

#ifdef __i386__
		fixup_guest_code_selector(c->event_callback_cs);
		fixup_guest_code_selector(c->failsafe_callback_cs);
#endif

		for ( i = 0; i < 256; i++ )
			fixup_guest_code_selector(c->trap_ctxt[i].cs);
	}
	else if ( !hvm_enabled )
		return -EINVAL;

	clear_bit(_VCPUF_fpu_initialised, &v->vcpu_flags);
	if ( c->flags & VGCF_I387_VALID )
		set_bit(_VCPUF_fpu_initialised, &v->vcpu_flags);

	v->arch.flags &= ~TF_kernel_mode;
	if ( (c->flags & VGCF_IN_KERNEL) || (c->flags & VGCF_HVM_GUEST) )
		v->arch.flags |= TF_kernel_mode;

	memcpy(&v->arch.guest_context, c, sizeof(*c));
	init_int80_direct_trap(v);

	if ( !(c->flags & VGCF_HVM_GUEST) )
	{
		/* IOPL privileges are virtualised. */
		v->arch.iopl = (v->arch.guest_context.user_regs.eflags >> 12) & 3;
		v->arch.guest_context.user_regs.eflags &= ~EF_IOPL;

		/* Ensure real hardware interrupts are enabled. */
		v->arch.guest_context.user_regs.eflags |= EF_IE;
	}
	else if ( test_bit(_VCPUF_initialised, &v->vcpu_flags) )
	{
		hvm_load_cpu_guest_regs(v, &v->arch.guest_context.user_regs);
	}

	if ( test_bit(_VCPUF_initialised, &v->vcpu_flags) )
		return 0;

	memset(v->arch.guest_context.debugreg, 0,
	       sizeof(v->arch.guest_context.debugreg));
	for ( i = 0; i < 8; i++ )
		(void)set_debugreg(v, i, c->debugreg[i]);

	if ( v->vcpu_id == 0 )
		d->vm_assist = c->vm_assist;

	if ( !(c->flags & VGCF_HVM_GUEST) )
	{
		phys_basetab = c->ctrlreg[3];
		phys_basetab =
			(gmfn_to_mfn(d, phys_basetab >> PAGE_SHIFT) << PAGE_SHIFT) |
			(phys_basetab & ~PAGE_MASK);

		v->arch.guest_table = mk_pagetable(phys_basetab);
	}

	if ( (rc = (int)set_gdt(v, c->gdt_frames, c->gdt_ents)) != 0 )
		return rc;

	if ( c->flags & VGCF_HVM_GUEST )
	{
		v->arch.guest_table = mk_pagetable(0);

		if ( !hvm_initialize_guest_resources(v) )
			return -EINVAL;
	}
	else if ( shadow_mode_refcounts(d) )
	{
		if ( !get_page(mfn_to_page(phys_basetab>>PAGE_SHIFT), d) )
		{
			destroy_gdt(v);
			return -EINVAL;
		}
	}
	else
	{
		if ( !get_page_and_type(mfn_to_page(phys_basetab>>PAGE_SHIFT), d,
					PGT_base_page_table) )
		{
			destroy_gdt(v);
			return -EINVAL;
		}
	}

	update_pagetables(v);

	if ( v->vcpu_id == 0 )
		init_domain_time(d);

	/* Don't redo final setup */
	set_bit(_VCPUF_initialised, &v->vcpu_flags);

	return 0;
}

void
vcpu_boot ( struct vcpu *vcpu, struct domain *dom );
{
	vcpu->domain = dom;
	
	struct vcpu_guest_context *ctxt = ctxt = xmalloc ( struct vcpu_guest_context );
	copy_from_guest ( ctxt, arg, 1 ); // ??? vmm operation ΰ
	arch_set_info_guest ( v, ctxt );
	xfree ( ctxt );
	
	arch_svm_do_launch ( vcpu, domain );
}

